import pytest
from group_lasso import _subsampling
import numpy as np


@pytest.fixture
def row_lengths():
    def _row_lengths():
        for i in range(20):
            yield 2 ** i - 1

    return _row_lengths


def test_random_row_idxes_correct_size_fraction(row_lengths):
    for row_length in row_lengths():
        for fraction in [0.5, 0.1, 1 / np.sqrt(2)]:
            row_idxes = _subsampling._get_random_row_idxes(row_length, fraction)
            assert len(row_idxes) == int(row_length) * fraction


def test_random_row_idxes_correct_size_fraction(row_lengths):
    for row_length in row_lengths():
        for num in [1, 10, 1000]:
            if num > row_length:
                continue
            row_idxes = _subsampling._get_random_row_idxes(row_length, num)
            assert len(row_idxes) == num


def test_random_row_idxes_fails_at_negative_input(row_lengths):
    for row_length in row_lengths():
        with pytest.raises(ValueError):
            row_idxes = _subsampling._get_random_row_idxes(row_length, -0.1)
        with pytest.raises(ValueError):
            row_idxes = _subsampling._get_random_row_idxes(row_length, -1)
        with pytest.raises(ValueError):
            row_idxes = _subsampling._get_random_row_idxes(row_length, 0)
        with pytest.raises(ValueError):
            row_idxes = _subsampling._get_random_row_idxes(row_length, row_length + 1)
        with pytest.raises(ValueError):
            row_idxes = _subsampling._get_random_row_idxes(row_length, "invalid")


def test_random_row_idxes_sqrt(row_lengths):
    for row_length in row_lengths():
        row_idxes = _subsampling._get_random_row_idxes(row_length, "sqrt")
        assert len(row_idxes) == int(np.sqrt(row_length))


def test_random_row_idxes_unique(row_lengths):
    for row_length in row_lengths():
        row_idxes = _subsampling._get_random_row_idxes(row_length, "sqrt")
        assert len(row_idxes) == len(set(row_idxes))



import numpy.linalg as la
from math import sqrt
import warnings


def _fista_momentum(momentum):
    return 0.5 + 0.5 * sqrt(1 + 4 * momentum ** 2)


def _fista_it(x, momentum_x, momentum, lipschitz, grad, prox):
    new_x = prox(momentum_x - grad(momentum_x) / lipschitz)
    new_momentum = _fista_momentum(momentum)

    dx = new_x - x
    new_momentum_x = new_x + dx * (momentum - 1) / momentum

    if (momentum_x.ravel() - new_x.ravel()).T @ (new_x.ravel() - x.ravel()) > 0:
        new_x, new_momentum_x, new_momentum = _fista_it(x, x, 1, lipschitz, grad, prox)

    return new_x, new_momentum_x, new_momentum


def fista(x0, grad, prox, loss, lipschitz, n_iter=10, tol=1e-6, callback=None):
    """Use the FISTA algorithm to solve the given optimisation problem
    """
    if callback is not None:
        callback(x0, 0)

    optimal_x = x0
    momentum_x = x0
    momentum = 1

    for i in range(n_iter):
        previous_x = optimal_x
        optimal_x, momentum_x, momentum = _fista_it(
            optimal_x, momentum_x, momentum, lipschitz, grad, prox
        )

        if callback is not None:
            callback(optimal_x, i, previous_x=previous_x)

        if la.norm(optimal_x - previous_x) / la.norm(optimal_x + 1e-16) < tol:
            return optimal_x

    warnings.warn(
        "The FISTA iterations did not converge to a sufficient minimum.\n"
        f"You used subsampling then this is expected, otherwise,"
        "try to increase the number of iterations "
        "or decreasing the tolerance.",
        RuntimeWarning,
    )

    return optimal_x



import numpy as np


def _extract_from_singleton_iterable(inputs):
    if len(inputs) == 1:
        return inputs[0]
    return tuple(inputs)


def _get_random_row_idxes(num_rows, subsampling_scheme):
    if subsampling_scheme is None:
        return range(num_rows)
    elif isinstance(subsampling_scheme, str):
        if subsampling_scheme.lower() == "sqrt":
            num_subsampled_rows = int(np.sqrt(num_rows))
        else:
            raise ValueError("Not valid subsampling scheme")
    elif subsampling_scheme < 1 and subsampling_scheme > 0:
        num_subsampled_rows = int(num_rows * subsampling_scheme)
    elif subsampling_scheme >= 1 and isinstance(subsampling_scheme, int):
        if subsampling_scheme > num_rows:
            raise ValueError("Cannot subsample more rows than there are present")
        num_subsampled_rows = subsampling_scheme
    else:
        raise ValueError("Not valid subsampling scheme")

    inds = np.random.choice(num_rows, num_subsampled_rows, replace=False)
    inds.sort()
    return inds


def subsampling_fraction(num_rows, subsampling_scheme):
    return len(_get_random_row_idxes(num_rows, subsampling_scheme)) / num_rows


def subsample(subsampling_scheme, *Xs):
    """Subsample along first (0-th) axis of the Xs arrays.

    Arguments
    ---------
    subsampling_scheme : int, float or str
        How to subsample:
         * int or float == 1 -> no subsampling
         * int > 1 -> that many rows are sampled
         * float < 1 -> the fraction of rows to subsample
         * sqrt -> subsample sqrt(num_rows) rows

    """
    assert len(Xs) > 0
    if subsampling_scheme == 1:
        return _extract_from_singleton_iterable(Xs)

    num_rows = len(Xs[0])
    inds = _get_random_row_idxes(num_rows, subsampling_scheme)
    return _extract_from_singleton_iterable([X[inds, :] for X in Xs])



from abc import ABC, abstractmethod
from math import sqrt
from numbers import Number
import warnings

import numpy.linalg as la
import numpy as np

from group_lasso._singular_values import find_largest_singular_value
from group_lasso._subsampling import subsample
from group_lasso._fista import fista


_DEBUG = False


def _l2_prox(w, reg):
    """The proximal operator for reg*||w||_2 (not squared).
    """
    return max(0, 1 - reg / la.norm(w)) * w


def _group_l2_prox(w, reg_coeffs, groups):
    """The proximal map for the specified groups of coefficients.
    """
    w = w.copy()

    for group, reg in zip(groups, reg_coeffs):
        reg = reg * sqrt(group.sum())
        w[group] = _l2_prox(w[group], reg)

    return w


def _split_intercept(w):
    return w[0], w[1:]


def _join_intercept(b, w):
    m, n = w.shape
    return np.concatenate([np.array(b).reshape(1, n), w], axis=0)


def _add_intercept_col(X):
    ones = np.ones([X.shape[0], 1])
    return np.concatenate([ones, X], axis=1)


class BaseGroupLasso(ABC):
    """
    This class implements the Group Lasso [1] regularisation for optimisation
    problems with Lipschitz continuous gradients, which is approximately
    equivalent to having a bounded third derivative.

    The loss is optimised using the FISTA algorithm proposed in [2] with the
    generalised gradient-based restarting scheme proposed in [3].

    [1]: Yuan M, Lin Y. Model selection and estimation in regression with
         grouped variables. Journal of the Royal Statistical Society: Series B
         (Statistical Methodology). 2006 Feb;68(1):49-67.
    [2]: Beck A, Teboulle M. A fast iterative shrinkage-thresholding algorithm
         for linear inverse problems. SIAM journal on imaging sciences.
         2009 Mar 4;2(1):183-202.
    [3]: O’donoghue B, Candes E. Adaptive restart for accelerated gradient
         schemes. Foundations of computational mathematics.
         2015 Jun 1;15(3):715-32.
    """

    # TODO: Document code
    # TODO: Use sklean mixins?
    # TODO: Tests

    def __init__(
        self,
        groups=None,
        reg=0.05,
        n_iter=1000,
        tol=1e-5,
        subsampling_scheme=None,
        fit_intercept=True,
    ):
        """

        Arguments
        ---------
        groups : Iterable
            Iterable that specifies which group each column corresponds to.
            For columns that should not be regularised, the corresponding
            group index should either be None or negative. For example, the
            list ``[1, 1, 1, 2, 2, -1]`` specifies that the first three
            columns of the data matrix belong to the first group, the next
            two columns belong to the second group and the last column should
            not be regularised.
        reg : float or iterable
            The regularisation coefficient(s). If ``reg`` is an
            iterable, then it should have the same length as
            ``groups``.
        n_iter : int
            The maximum number of iterations to perform
        tol : float
            The convergence tolerance. The optimisation algorithm
            will stop once ||x_{n+1} - x_n|| < ``tol``.
        subsampling_scheme : float, int or str
            The subsampling rate used for the gradient and singular value
            computations. If it is a float, then it specifies the fraction
            of rows to use in the computations. If it is an int, it
            specifies the number of rows to use in the computation and if
            it is a string, then it must be 'sqrt' and the number of rows used
            in the computations is the square root of the number of rows
            in X.
        fit_intercept : bool (default=True)
        """
        self.groups = groups
        self.reg = reg
        self.n_iter = n_iter
        self.tol = tol
        self.subsampling_scheme = subsampling_scheme
        self.fit_intercept = fit_intercept

    def get_params(self, deep=True):
        return {
            "groups": self.groups,
            "reg": self.reg,
            "n_iter": self.n_iter,
            "tol": self.tol,
            "subsampling_scheme": self.subsampling_scheme,
        }

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

    def _regularizer(self, w):
        regularizer = 0
        b, w = _split_intercept(w)
        for group, reg in zip(self.groups_, self.reg_vector):
            regularizer += reg * la.norm(w[group, :])
        return regularizer

    def _get_reg_vector(self, reg):
        if isinstance(reg, Number):
            return [reg * sqrt(group.sum()) for group in self.groups_]
        return reg

    @abstractmethod
    def _unregularised_loss(self, X, y, w):
        pass

    def _loss(self, X, y, w):
        return self._unregularised_loss(X, y, w) + self._regularizer(w)

    def loss(self, X, y):
        return self._loss(X, y, self.coef_)

    @abstractmethod
    def _compute_lipschitz(self, X, y):
        pass

    @abstractmethod
    def _grad(self, X, y, w):
        pass

    def _minimise_loss(self, X, y, lipschitz=None):
        """Use the FISTA algorithm to solve the group lasso regularised loss.
        """
        if self.fit_intercept:
            X = _add_intercept_col(X)

        if lipschitz is None:
            lipschitz = self._compute_lipschitz(X, y)

        if not self.fit_intercept:
            X = _add_intercept_col(X)

        def grad(w):
            g = self._grad(X, y, w)
            if not self.fit_intercept:
                g[0] = 0
            return g

        def prox(w):
            b, w_ = _split_intercept(w)
            w_ = _group_l2_prox(w_, self.reg_vector, self.groups_)
            return _join_intercept(b, w_)

        def loss(w):
            X_, y_ = subsample(self.subsampling_scheme, X, y)
            self._loss(X_, y_, w)

        def callback(x, it_num, previous_x=None):
            X_, y_ = subsample(self.subsampling_scheme, X, y)
            w = x
            previous_w = previous_x

            self.losses_.append(self._loss(X_, y_, w))

            if previous_w is None and _DEBUG:
                print(f"Starting FISTA: ")
                print(f"\tInitial loss: {self._loss(X_, y_, w)}")

            elif _DEBUG:
                print(f"Completed iteration {it_num}:")
                print(f"\tLoss: {self._loss(X_, y_, w)}")
                print(f"\tWeight difference: {la.norm(w-previous_w)}")
                print(f"\tWeight norm: {la.norm(w)}")
                print(f"\tGrad: {la.norm(grad(w))}")

        weights = np.concatenate([self.intercept_, self.coef_])
        weights = fista(
            weights,
            grad=grad,
            prox=prox,
            loss=loss,
            lipschitz=lipschitz,
            n_iter=self.n_iter,
            tol=self.tol,
            callback=callback,
        )
        self.intercept_, self.coef_ = _split_intercept(weights)

        warnings.warn(
            "The FISTA iterations did not converge to a sufficient minimum.\n"
            f"You used subsampling then this is expected, otherwise,"
            "try to increase the number of iterations "
            "or decreasing the tolerance.",
            RuntimeWarning,
        )

    def _check_valid_parameters(self):
        assert all(reg >= 0 for reg in self.reg_vector)
        assert len(self.reg_vector) == len(np.unique(self.groups))
        assert self.n_iter > 0
        assert self.tol >= 0

    def _prepare_dataset(self, X, y):
        if len(y.shape) != 1:
            pass
            # assert y.shape[1] == 1
        else:
            y = y.reshape(-1, 1)
        return X, y

    def _init_fit(self, X, y):
        X, y = self._prepare_dataset(X, y)
        groups = np.array([-1 if i is None else i for i in self.groups])
        self.groups_ = [self.groups == u for u in np.unique(groups) if u >= 0]
        self.reg_vector = self._get_reg_vector(self.reg)
        self.losses_ = []
        self.coef_ = np.random.randn(X.shape[1], y.shape[1])
        self.coef_ /= la.norm(self.coef_)
        self.intercept_ = np.zeros((1, self.coef_.shape[1]))

        self._check_valid_parameters()

    def fit(self, X, y, lipschitz=None):
        self._init_fit(X, y)
        self._minimise_loss(X, y, lipschitz=lipschitz)

    @abstractmethod
    def predict(self, X):
        pass

    def fit_predict(self, X, y):
        self.fit(X, y)
        return self.predict(X)

    @property
    def sparsity_mask(self):
        pattern = np.zeros(len(self.coef_), dtype=bool)
        for group in self.groups_:
            pattern[group] = not np.allclose(self.coef_[group], 0, atol=0, rtol=1e-10)

        return pattern

    def transform(self, X):
        return X[:, self.sparsity_mask]

    def fit_transform(self, X, y, lipschitz=None):
        self.fit(X, y, lipschitz)
        return self.transform(X)

    def subsample(self, *args):
        return subsample(self.subsampling_scheme, *args)


def _l2_grad(A, b, x):
    """The gradient of the problem ||Ax - b||^2 wrt x.
    """
    return A.T @ (A @ x - b)


class GroupLasso(BaseGroupLasso):
    """
    This class implements the Group Lasso [1] regularisation for linear
    regression with the mean squared penalty.

    This class is implemented as both a regressor and a transformation.
    If the ``transform`` method is called, then the columns of the input
    that correspond to zero-valued regression coefficients are dropped.

    The loss is optimised using the FISTA algorithm proposed in [2] with the
    generalised gradient-based restarting scheme proposed in [3]. This
    algorithm is not as accurate as a few other optimisation algorithms,
    but it is extremely efficient and does recover the sparsity patterns.
    We therefore reccomend that this class is used as a transformer to select
    the viable features and that the output is fed into another regression
    algorithm, such as RidgeRegression in scikit-learn.

    References
    ----------
    [1]: Yuan M, Lin Y. Model selection and estimation in regression with
         grouped variables. Journal of the Royal Statistical Society: Series B
         (Statistical Methodology). 2006 Feb;68(1):49-67.
    [2]: Beck A, Teboulle M. A fast iterative shrinkage-thresholding algorithm
         for linear inverse problems. SIAM journal on imaging sciences.
         2009 Mar 4;2(1):183-202.
    [3]: O’donoghue B, Candes E. Adaptive restart for accelerated gradient
         schemes. Foundations of computational mathematics.
         2015 Jun 1;15(3):715-32.
    """

    def __init__(
        self,
        groups=None,
        reg=0.05,
        n_iter=1000,
        tol=1e-5,
        subsampling_scheme=None,
        fit_intercept=True,
        frobenius_lipschitz=False,
    ):
        """

        Arguments
        ---------
        groups : Iterable
            Iterable that specifies which group each column corresponds to.
            For columns that should not be regularised, the corresponding
            group index should either be None or negative. For example, the
            list ``[1, 1, 1, 2, 2, -1]`` specifies that the first three
            columns of the data matrix belong to the first group, the next
            two columns belong to the second group and the last column should
            not be regularised.
        reg : float or iterable
            The regularisation coefficient(s). If ``reg`` is an
            iterable, then it should have the same length as
            ``groups``.
        n_iter : int
            The maximum number of iterations to perform
        tol : float
            The convergence tolerance. The optimisation algorithm
            will stop once ||x_{n+1} - x_n|| < ``tol``.
        subsampling_scheme : float, int or str
            The subsampling rate used for the gradient and singular value
            computations. If it is a float, then it specifies the fraction
            of rows to use in the computations. If it is an int, it
            specifies the number of rows to use in the computation and if
            it is a string, then it must be 'sqrt' and the number of rows used
            in the computations is the square root of the number of rows
            in X.
        frobenius_lipschitz : Bool [default=False]
            Use the Frobenius norm to estimate the lipschitz coefficient of the
            MSE loss. This works well for systems whose power iterations
            converge slowly. If False, then subsampled power iterations are
            used. Using the Frobenius approximation for the Lipschitz
            coefficient might fail, and end up with all-zero weights.
        fit_intercept : bool (default=True)
        """
        super().__init__(
            groups=groups,
            reg=reg,
            n_iter=n_iter,
            tol=tol,
            subsampling_scheme=subsampling_scheme,
            fit_intercept=fit_intercept,
        )
        self.frobenius_lipchitz = frobenius_lipschitz

    def predict(self, X):
        return self.intercept_ + X @ self.coef_

    def _unregularised_loss(self, X, y, w):
        X_, y_ = self.subsample(X, y)
        MSE = 0.5 * np.sum((X_ @ w - y_) ** 2) / len(X_)
        return MSE

    def _grad(self, X, y, w):
        X_, y_ = self.subsample(X, y)
        SSE_grad = _l2_grad(X_, y_, w)
        return SSE_grad / len(X_)

    def _compute_lipschitz(self, X, y):
        num_rows, num_cols = X.shape
        if self.frobenius_lipchitz:
            return la.norm(X, "fro") ** 2 / (num_rows * num_cols)

        SSE_lipschitz = (
            1.5
            * find_largest_singular_value(X, subsampling_scheme=self.subsampling_scheme)
            ** 2
        )
        return SSE_lipschitz / num_rows


def _sigmoid(x):
    return 1 / (1 + np.exp(-x))


def _logit(X, w):
    return X @ w


def _logistic_proba(X, w):
    return _sigmoid(_logit(X, w))


def _logistic_cross_entropy(X, y, w):
    p = _logistic_proba(X, w)
    return -(y * np.log(p) + (1 - y) * np.log(1 - p))


class LogisticGroupLasso(BaseGroupLasso):
    """WARNING: Experimental.
    """

    def _compute_proba(self, X, w):
        return _sigmoid(X @ w)

    def _unregularised_loss(self, X, y, w):
        X_, y_ = self.subsample(X, y)
        return _logistic_cross_entropy(X_, y_, w).sum() / len(X)

    def _grad(self, X, y, w):
        X_, y_ = self.subsample(X, y)
        p = _logistic_proba(X_, w)
        return X_.T @ (p - y_) / len(X_)

    def _compute_lipschitz(self, X, y):
        return np.sqrt(12) * np.linalg.norm(X, "fro") / len(X)

    def predict_proba(self, X):
        return _logistic_proba(X, self.coef_)

    def predict(self, X):
        return self.predict_proba(X) >= 0.5

    def fit(self, X, y, lipschitz=None):
        if y.ndim == 2 and y.shape[1] > 1:
            n = y.shape[1]
            warnings.warn(
                f"You have passed {n} targets to a single class classifier. "
                f"This will simply train {n} different models meaning that "
                f"multiple classes can be predicted as true at once."
            )

        super().fit(X, y, lipschitz=lipschitz)



from setuptools import setup
from setuptools import find_packages


with open("README.md") as f:
    long_description = f.read()


setup(
    name="group-lasso",
    version="0.1.0",
    license="MIT",
    description="Fast group lasso regularised linear models in a sklearn-style API.",
    long_description=long_description,
    author="Yngve Mardal Moe",
    author_email="yngve.m.moe@gmail.com",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        'numpy',
        'scikit-learn'
    ],
    extras_require={'test': ['pytest']}
)



import warnings

import numpy as np
import numpy.linalg as la
from math import sqrt

from group_lasso._subsampling import subsample, subsampling_fraction

_DEBUG = False
LIPSCHITZ_MAXITS = 20
LIPSCHITS_TOL = 5e-3


def _power_iteration(X, v):
    v = X.T @ (X @ v)
    s = la.norm(v)
    v /= s
    return v, s


def _subsampled_power_iteration(X, v, subsampling_scheme):
    X_ = subsample(subsampling_scheme, X)
    v, s = _power_iteration(X_, v)

    return v, s / subsampling_fraction(len(X), subsampling_scheme)


def find_largest_singular_value(
    X, subsampling_scheme=None, maxits=LIPSCHITZ_MAXITS, tol=LIPSCHITS_TOL
):
    """Find the largest singular value of X.
    """
    v = np.random.randn(X.shape[1], 1)
    s = la.norm(v)
    v /= s
    for i in range(maxits):
        s_ = s
        v, s = _subsampled_power_iteration(X, v, subsampling_scheme)

        # Absolute value is necessary because of subsampling
        improvement = abs(s - s_) / max(abs(s), abs(s_))
        if improvement < tol and i > 0:
            return np.sqrt(s)

        if _DEBUG:
            print(
                f"Finished {i}th power iteration:\n"
                f"\tL={sqrt(s)}\n"
                f"\tImprovement: {improvement:03g}"
            )

    warnings.warn(
        f"Could not find an estimate for the largest singular value of X"
        f"with the power method. \n"
        f"Ran for {maxits:d} iterations with a tolerance of {tol:02g}"
        f'Subsampling {"is" if subsampling_scheme is not None else "is not"} used.',
        RuntimeWarning,
    )
    return np.sqrt(s)



import numpy as np
import numpy.linalg as la
import pytest

from group_lasso import _singular_values


np.random.seed(0)
TOL = 0.1
SUBSAMPLED_TOL = 0.1
NUM_COLS, NUM_ROWS = 1000, 50
LARGE_COLS, LARGE_ROWS = 10000, 1000
_singular_values._DEBUG = True


@pytest.fixture
def random_svd():
    X = np.random.randn(NUM_COLS, NUM_ROWS)
    U, s, Vh = la.svd(X)
    s = s ** 2
    return U, s, Vh


@pytest.fixture
def large_random_svd():
    X = np.random.randn(LARGE_COLS, LARGE_ROWS)
    U, s, Vh = la.svd(X, full_matrices=False)
    s = s ** 2
    return U, s, Vh


def generate_matrix_from_svd(U, s, Vh):
    return (U[:, : len(s)] * s) @ Vh[: len(s)]


def test_find_largest_singular_value(random_svd):
    s = random_svd[1]
    X = generate_matrix_from_svd(*random_svd)
    smax = _singular_values.find_largest_singular_value(X)

    assert abs(smax - s[0]) / max(smax, s[0]) < TOL
    assert smax < s[0]


def test_power_iteration(random_svd):
    X = generate_matrix_from_svd(*random_svd)
    v0 = np.random.randn(NUM_ROWS)

    v1, s1 = _singular_values._power_iteration(X, v0)
    v1_ = X.T @ (X @ v0)
    s1_ = la.norm(v1_)
    v1_ /= s1_

    assert abs(s1_ - s1) < 1e-10
    assert np.allclose(v1, v1_)


def test_subsampled_find_largest_singular_value(random_svd):
    s = random_svd[1]
    X = generate_matrix_from_svd(*random_svd)
    smax = _singular_values.find_largest_singular_value(X, subsampling_scheme="sqrt")

    assert abs(smax - s[0]) / max(smax, s[0]) < SUBSAMPLED_TOL



from contextlib import contextmanager
import numpy as np
import numpy.linalg as la
import pytest
from group_lasso import _group_lasso
from sklearn.linear_model import LinearRegression, LogisticRegression

np.random.seed(0)


class BaseTestGroupLasso:
    MLFitter = _group_lasso.BaseGroupLasso
    UnregularisedMLFitter = None
    num_rows = 200
    num_cols = 30
    configs = [{"n_iter": 1000}]

    @contextmanager
    def all_configs(self, gl):
        for config in self.configs:
            gl.set_params(**config, tol=0)
            yield gl

    @pytest.fixture
    def gl_no_reg(self):
        return self.MLFitter(reg=0, groups=[])

    @pytest.fixture
    def sklearn_no_reg(self):
        return self.UnregularisedMLFitter()

    @pytest.fixture
    def ml_problem(self):
        raise NotImplementedError

    def random_weights(self):
        return np.random.standard_normal((self.num_cols, 1))

    def test_lipschits(self, gl_no_reg, ml_problem):
        X, y, w = ml_problem

        with self.all_configs(gl_no_reg) as gl:
            L = gl._compute_lipschitz(X, y)

            g1 = gl._grad(X, y, w)
            for i in range(100):
                w2 = self.random_weights() * i

                g2 = gl._grad(X, y, w2)

                assert la.norm(g1 - g2) <= L * la.norm(w - w2)

    def test_grad(self, gl_no_reg, ml_problem):
        X, y, w = ml_problem
        w = self.random_weights()
        eps = 1e-10

        with self.all_configs(gl_no_reg) as gl:
            loss = gl._unregularised_loss(X, y, w)
            print(loss)
            dw = np.empty_like(w)
            for i, _ in enumerate(w):
                w_ = w.copy()
                w_[i] += eps
                dw[i] = (gl._unregularised_loss(X, y, w_) - loss) / (w_[i] - w[i])
                g = gl._grad(X, y, w)
            assert la.norm(dw - g) / la.norm(g) < 1e-3

    def test_unregularised_fit_equal_sklearn(
        self, gl_no_reg, sklearn_no_reg, ml_problem
    ):
        X, y, w = ml_problem
        with self.all_configs(gl_no_reg) as gl:
            yhat1 = gl.fit_predict(X, y)
            sklearn_no_reg.fit(X, y)
            yhat2 = sklearn_no_reg.predict(X).reshape(yhat1.shape)

            assert np.allclose(yhat1, yhat2)

    # TODO: FIND SPARSITY PATTERNS WITH NOISE


class TestGroupLasso(BaseTestGroupLasso):
    MLFitter = _group_lasso.GroupLasso
    UnregularisedMLFitter = LinearRegression

    @pytest.fixture
    def ml_problem(self):
        X = np.random.standard_normal((self.num_rows, self.num_cols))
        w = self.random_weights()
        y = X @ w
        return X, y, w


class TestLogisticGroupLasso(BaseTestGroupLasso):
    MLFitter = _group_lasso.LogisticGroupLasso
    UnregularisedMLFitter = LogisticRegression

    @pytest.fixture
    def ml_problem(self):
        X = np.random.standard_normal((self.num_rows, self.num_cols))
        w = self.random_weights()
        y = _group_lasso._sigmoid(X @ w) > 0.5
        return X, y, w

    def test_unregularised_fit_equal_sklearn(
        self, gl_no_reg, sklearn_no_reg, ml_problem
    ):
        X, y, w = ml_problem
        with self.all_configs(gl_no_reg) as gl:
            yhat1 = gl.fit_predict(X, y)
            sklearn_no_reg.fit(X, y)
            yhat2 = sklearn_no_reg.predict(X)

            assert np.mean(yhat1.astype(float) - yhat2.astype(float)) < 1e-2
